from typing import Optional

import aiohttp

from esihub.core.logger import esihub_logger


class ESIConnectionPool:
    def __init__(self, pool_size: int = 100):
        self.pool_size = pool_size
        self.session: Optional[aiohttp.ClientSession] = None
        self.logger = esihub_logger

    async def get(self) -> aiohttp.ClientSession:
        if self.session is None or self.session.closed:
            self.session = aiohttp.ClientSession(
                connector=aiohttp.TCPConnector(limit=self.pool_size)
            )
            self.logger.debug(
                f"Created new ClientSession with pool size {self.pool_size}"
            )
        return self.session

    async def release(self, session: aiohttp.ClientSession) -> None:
        if session and not session.closed:
            await session.close()
            self.logger.debug("Closed ClientSession")

    async def close(self) -> None:
        if self.session and not self.session.closed:
            await self.session.close()
            self.logger.debug("Closed ConnectionPool")



import asyncio
import time
from typing import Dict

from multidict import CIMultiDictProxy

from .config import ESIHubConfig
from .logger import esihub_logger


class ESIHubRateLimiter:
    def __init__(self, config: ESIHubConfig):
        self.config = config
        self.limiters: Dict[str, Dict] = {}
        self.global_limit = self.config.get("ESI_RATE_LIMIT", 150)
        self.global_remaining = self.global_limit
        self.global_reset = 0
        self.lock = asyncio.Lock()

    async def acquire(self, endpoint: str):
        async with self.lock:
            await self._check_global_limit()

            if endpoint not in self.limiters:
                self.limiters[endpoint] = {
                    "limit": self.global_limit,
                    "remaining": self.global_limit,
                    "reset": 0,
                }

            limiter = self.limiters[endpoint]
            current_time = time.time()

            if current_time > limiter["reset"]:
                limiter["remaining"] = limiter["limit"]
                limiter["reset"] = current_time + 1  # Reset every second

            if limiter["remaining"] <= 0:
                wait_time = limiter["reset"] - current_time
                esihub_logger.debug(
                    "Rate limit reached, waiting",
                    extra={"endpoint": endpoint, "wait_time": wait_time},
                )
                await asyncio.sleep(wait_time)
                limiter["remaining"] = limiter["limit"]
                limiter["reset"] = time.time() + 1

            limiter["remaining"] -= 1
            self.global_remaining -= 1

    async def _check_global_limit(self):
        current_time = time.time()
        if current_time > self.global_reset:
            self.global_remaining = self.global_limit
            self.global_reset = current_time + 1

        if self.global_remaining <= 0:
            wait_time = self.global_reset - current_time
            esihub_logger.debug(
                "Global rate limit reached, waiting", extra={"wait_time": wait_time}
            )
            await asyncio.sleep(wait_time)
            self.global_remaining = self.global_limit
            self.global_reset = time.time() + 1

    def update_limit(self, endpoint: str, headers: CIMultiDictProxy[str]):
        if (
            "X-Esi-Error-Limit-Remain" in headers
            and "X-Esi-Error-Limit-Reset" in headers
        ):
            remain = int(headers["X-Esi-Error-Limit-Remain"])
            reset = int(headers["X-Esi-Error-Limit-Reset"])

            if endpoint in self.limiters:
                self.limiters[endpoint]["remaining"] = remain
                self.limiters[endpoint]["reset"] = time.time() + reset
            else:
                self.limiters[endpoint] = {
                    "limit": self.global_limit,
                    "remaining": remain,
                    "reset": time.time() + reset,
                }

        if "X-Esi-Error-Limit-Remain" in headers:
            self.global_remaining = int(headers["X-Esi-Error-Limit-Remain"])
        if "X-Esi-Error-Limit-Reset" in headers:
            self.global_reset = time.time() + int(headers["X-Esi-Error-Limit-Reset"])

        esihub_logger.debug(
            "Rate limit updated",
            extra={
                "endpoint": endpoint,
                "remaining": self.global_remaining,
                "reset": self.global_reset,
            },
        )



import asyncio
from typing import Callable, Awaitable, Any, List

from .logger import esihub_logger


class ESIHubBackgroundTaskManager:
    def __init__(self):
        self.tasks: List[asyncio.Task] = []
        self.queue: asyncio.PriorityQueue = asyncio.PriorityQueue()
        self.running = False

    async def start(self):
        self.running = True
        self.process_queue_task = asyncio.create_task(self._process_queue())

    async def stop(self):
        self.running = False
        # Cancel all running tasks
        for task in self.tasks:
            task.cancel()
        # Await their completion to ensure they are properly cleaned up
        await asyncio.gather(*self.tasks, return_exceptions=True)
        # Ensure process_queue_task is also cancelled and awaited
        self.process_queue_task.cancel()
        await asyncio.gather(self.process_queue_task, return_exceptions=True)

    async def add_task(
        self,
        coroutine: Callable[..., Awaitable[Any]],
        *args,
        priority: int = 0,
        **kwargs,
    ):
        await self.queue.put((-priority, coroutine, args, kwargs))

    async def _process_queue(self):
        while self.running:
            try:
                _, coroutine, args, kwargs = await self.queue.get()
                task = asyncio.create_task(coroutine(*args, **kwargs))
                self.tasks.append(task)
                task.add_done_callback(self._task_done)
            except asyncio.CancelledError:
                break
            except Exception as e:
                esihub_logger.error(f"Error processing background task: {str(e)}")
        # Cleanup when stopping
        while not self.queue.empty():
            try:
                self.queue.get_nowait()
            except asyncio.QueueEmpty:
                break

    def _task_done(self, task: asyncio.Task):
        if task in self.tasks:
            self.tasks.remove(task)
        try:
            task.result()
        except asyncio.CancelledError:
            pass
        except Exception as e:
            esihub_logger.error(f"Background task failed: {str(e)}")



import json
import logging
from typing import Any, Dict


class ESIHubLogger:
    def __init__(self, name: str, level: int = logging.INFO):
        self.logger = logging.getLogger(name)
        self.set_level(level)
        handler = logging.StreamHandler()
        formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )
        handler.setFormatter(formatter)
        self.logger.addHandler(handler)

    def set_level(self, level: int):
        self.logger.setLevel(level)

    def _format_log(self, message: str, extra: Dict[str, Any] = None) -> str:
        log_data = {"message": message}
        if extra:
            log_data.update(extra)
        return json.dumps(log_data)

    def debug(self, message: str, extra: Dict[str, Any] = None):
        self.logger.debug(self._format_log(message, extra))

    def info(self, message: str, extra: Dict[str, Any] = None):
        self.logger.info(self._format_log(message, extra))

    def warning(self, message: str, extra: Dict[str, Any] = None):
        self.logger.warning(self._format_log(message, extra))

    def error(self, message: str, extra: Dict[str, Any] = None, exc_info: bool = False):
        self.logger.error(self._format_log(message, extra), exc_info=exc_info)

    def critical(
        self, message: str, extra: Dict[str, Any] = None, exc_info: bool = False
    ):
        self.logger.critical(self._format_log(message, extra), exc_info=exc_info)


esihub_logger = ESIHubLogger("esihub")


def configure_logging(config):
    log_level = getattr(logging, config.get("LOG_LEVEL", "INFO").upper())
    esihub_logger.set_level(log_level)



from typing import Dict, Any, Optional

from pydantic import BaseModel, Field


class ESIHubRequestParams(BaseModel):
    method: str
    path: str
    query_params: Optional[Dict[str, Any]] = Field(default_factory=dict)
    body: Optional[Dict[str, Any]] = None
    headers: Optional[Dict[str, str]] = Field(default_factory=dict)


class ESIHubResponse(BaseModel):
    status: int
    headers: Dict[str, str]
    data: Any



import asyncio
import json
import os
from functools import wraps
from typing import Dict, Any, Callable

from .core.logger import esihub_logger


def load_swagger_spec(file_path: str = None) -> Dict[str, Any]:
    if file_path is None:
        file_path = os.path.join(os.path.dirname(__file__), "swagger.json")

    try:
        with open(file_path, "r") as f:
            return json.load(f)
    except FileNotFoundError:
        esihub_logger.error(f"Swagger spec file not found: {file_path}")
        raise
    except json.JSONDecodeError:
        esihub_logger.error(f"Invalid JSON in Swagger spec file: {file_path}")
        raise


def parse_path_params(path: str) -> Dict[str, Any]:
    params = {}
    parts = path.split("/")
    for part in parts:
        if part.startswith("{") and part.endswith("}"):
            param_name = part[1:-1]
            params[param_name] = None
    return params


def replace_path_params(path: str, **kwargs: Any) -> str:
    for key, value in kwargs.items():
        path = path.replace(f"{{{key}}}", str(value))
    return path


def retry_with_exponential_backoff(
    max_retries: int = 3, base_delay: float = 1, max_delay: float = 60
):
    def decorator(func: Callable):
        @wraps(func)
        async def wrapper(*args: Any, **kwargs: Any):
            retries = 0
            while True:
                try:
                    return await func(*args, **kwargs)
                except Exception as e:
                    retries += 1
                    if retries > max_retries:
                        raise
                    delay = min(base_delay * (2 ** (retries - 1)), max_delay)
                    esihub_logger.warning(
                        f"Retry {retries}/{max_retries} after {delay:.2f}s. Error: {str(e)}"
                    )
                    await asyncio.sleep(delay)

        return wrapper

    return decorator


def validate_url(url: str) -> bool:
    from urllib.parse import urlparse

    try:
        result = urlparse(url)
        return all([result.scheme, result.netloc])
    except Exception as e:
        esihub_logger.error(f"Error validating URL: {str(e)}")
        return False


def validate_input(input_str: str, pattern: str) -> bool:
    import re

    return bool(re.match(pattern, input_str))



from .client import ESIHubClient
from .exceptions import (
    ESIHubError,
    ESIHubAuthenticationError,
    ESIHubRateLimitError,
    ESIHubServerError,
    ESIHubClientError,
    ESIHubValidationError,
)
from .models import ESIHubRequestParams, ESIHubResponse

__all__ = [
    "ESIHubClient",
    "ESIHubError",
    "ESIHubAuthenticationError",
    "ESIHubRateLimitError",
    "ESIHubServerError",
    "ESIHubClientError",
    "ESIHubValidationError",
    "ESIHubRequestParams",
    "ESIHubResponse",
]

__version__ = "0.0.3"



from typing import Dict, Any
from urllib.parse import urlencode

import aiohttp

from .core.config import ESIHubConfig
from .exceptions import ESIHubAuthenticationError


class ESIHubAuth:
    def __init__(self, client, config: ESIHubConfig):
        self.client = client
        self.config = config

    async def get_auth_url(self, scopes: str = None, state: str = None) -> str:
        params = {
            "response_type": "code",
            "redirect_uri": self.config.get("ESI_CALLBACK_URL"),
            "client_id": self.config.get("ESI_CLIENT_ID"),
        }
        if scopes:
            params["scope"] = scopes
        if state:
            params["state"] = state
        return f"https://login.eveonline.com/v2/oauth/authorize?{urlencode(params)}"

    async def get_access_token(self, code: str) -> Dict[str, Any]:
        async with aiohttp.ClientSession() as session:
            async with session.post(
                "https://login.eveonline.com/v2/oauth/token",
                data={
                    "grant_type": "authorization_code",
                    "code": code,
                    "client_id": self.config.get("ESI_CLIENT_ID"),
                    "client_secret": self.config.get("ESI_CLIENT_SECRET"),
                },
            ) as resp:
                if resp.status != 200:
                    raise ESIHubAuthenticationError(
                        f"Failed to get access token: {await resp.text()}"
                    )
                return await resp.json()

    async def refresh_token(self, refresh_token: str) -> Dict[str, Any]:
        async with aiohttp.ClientSession() as session:
            async with session.post(
                "https://login.eveonline.com/v2/oauth/token",
                data={
                    "grant_type": "refresh_token",
                    "refresh_token": refresh_token,
                    "client_id": self.config.get("ESI_CLIENT_ID"),
                    "client_secret": self.config.get("ESI_CLIENT_SECRET"),
                },
            ) as resp:
                if resp.status != 200:
                    raise ESIHubAuthenticationError(
                        f"Failed to refresh token: {await resp.text()}"
                    )
                return await resp.json()



from typing import Dict, Any

from .logger import esihub_logger
from ..exceptions import (
    ESIHubAuthenticationError,
    ESIHubRateLimitError,
    ESIHubClientError,
    ESIHubServerError,
    ESIHubError,
)


class ESIHubErrorHandler:
    @staticmethod
    async def handle_error(status_code: int, response_data: Dict[str, Any]) -> None:
        error_message = response_data.get("error", "Unknown error")
        if 400 <= status_code < 500:
            if status_code == 401:
                raise ESIHubAuthenticationError(
                    f"Authentication failed: {error_message}",
                    status_code,
                    response_data,
                )
            elif status_code == 403:
                raise ESIHubAuthenticationError(
                    f"Insufficient permissions: {error_message}",
                    status_code,
                    response_data,
                )
            elif status_code == 420:
                raise ESIHubRateLimitError(
                    f"Rate limit exceeded: {error_message}", status_code, response_data
                )
            else:
                raise ESIHubClientError(
                    f"Client error: {error_message}", status_code, response_data
                )
        elif status_code >= 500:
            raise ESIHubServerError(
                f"Server error: {error_message}", status_code, response_data
            )
        else:
            raise ESIHubError(
                f"Unexpected error: {error_message}", status_code, response_data
            )

    @staticmethod
    def log_error(error: Exception) -> None:
        if isinstance(error, ESIHubError):
            esihub_logger.error(
                f"{error.__class__.__name__}: {str(error)}",
                extra={"status_code": error.status_code, "details": error.details},
            )
        else:
            esihub_logger.error(f"Unexpected error: {str(error)}", exc_info=True)



import time
from functools import wraps
from typing import Callable, Dict, List

from .logger import esihub_logger


def profile(func):
    @wraps(func)
    async def wrapper(*args, **kwargs):
        start_time = time.time()
        result = await func(*args, **kwargs)
        end_time = time.time()

        execution_time = end_time - start_time
        esihub_logger.info(
            f"Function {func.__name__} took {execution_time:.4f} seconds to execute"
        )

        return result

    return wrapper


class ESIHubAsyncProfiler:
    def __init__(self):
        self.function_times: Dict[str, List[float]] = {}

    async def profile_coroutine(self, coroutine):
        start_time = time.time()
        result = await coroutine
        end_time = time.time()

        execution_time = end_time - start_time
        coroutine_name = coroutine.__name__

        if coroutine_name not in self.function_times:
            self.function_times[coroutine_name] = []

        self.function_times[coroutine_name].append(execution_time)

        return result

    def print_stats(self):
        for func_name, times in self.function_times.items():
            avg_time = sum(times) / len(times)
            min_time = min(times)
            max_time = max(times)

            esihub_logger.info(f"Function: {func_name}")
            esihub_logger.info(f"  Calls: {len(times)}")
            esihub_logger.info(f"  Average time: {avg_time:.4f} seconds")
            esihub_logger.info(f"  Min time: {min_time:.4f} seconds")
            esihub_logger.info(f"  Max time: {max_time:.4f} seconds")

    async def run_profiled(self, func: Callable, *args, **kwargs):
        return await self.profile_coroutine(func(*args, **kwargs))



import time
from contextlib import contextmanager
from prometheus_client import CollectorRegistry, Counter, Histogram, Gauge


class ESIHubMetrics:
    def __init__(self):
        self.registry = CollectorRegistry()
        self.request_counter = self._get_or_create_counter(
            "esihub_requests_total", "Total requests made", ["method", "path"]
        )
        self.request_duration = self._get_or_create_histogram(
            "esihub_request_duration_seconds",
            "Request duration in seconds",
            ["method", "path"],
        )
        self.error_counter = self._get_or_create_counter(
            "esihub_errors_total", "Total errors encountered", ["error_type"]
        )
        self.active_requests = self._get_or_create_gauge(
            "esihub_active_requests", "Number of active requests"
        )

    def _get_or_create_counter(self, name, documentation, labelnames):
        try:
            return Counter(name, documentation, labelnames, registry=self.registry)
        except ValueError:
            return self.registry._names_to_collectors[name]

    def _get_or_create_histogram(self, name, documentation, labelnames):
        try:
            return Histogram(name, documentation, labelnames, registry=self.registry)
        except ValueError:
            return self.registry._names_to_collectors[name]

    def _get_or_create_gauge(self, name, documentation):
        try:
            return Gauge(name, documentation, registry=self.registry)
        except ValueError:
            return self.registry._names_to_collectors[name]

    def increment_request(self, method: str, path: str):
        self.request_counter.labels(method=method, path=path).inc()

    @contextmanager
    def measure_request_duration(self, method: str, path: str):
        start_time = time.time()
        self.active_requests.inc()
        try:
            yield
        finally:
            duration = time.time() - start_time
            self.request_duration.labels(method=method, path=path).observe(duration)
            self.active_requests.dec()

    def increment_error(self, error_type: str):
        self.error_counter.labels(error_type=error_type).inc()



from typing import Any, Callable, Coroutine, Dict

from esihub.core.logger import esihub_logger
from esihub.utils import parse_path_params, replace_path_params


def generate_endpoint_method(
    client, method: str, path: str, operation_id: str
) -> Callable[..., Coroutine[Any, Any, Dict[str, Any]]]:
    path_params = parse_path_params(path)

    async def endpoint_method(**kwargs) -> Dict[str, Any]:
        esihub_logger.debug(f"Calling endpoint: {operation_id}")
        path_args = {k: kwargs.pop(k) for k in path_params if k in kwargs}
        formatted_path = replace_path_params(path, **path_args)
        return await client.request(method=method, path=formatted_path, **kwargs)

    endpoint_method.__name__ = operation_id
    endpoint_method.__doc__ = f"{method.upper()} {path}"

    return endpoint_method


def generate_endpoints(client):
    for path, methods in client.swagger_spec["paths"].items():
        for method, details in methods.items():
            if "operationId" in details:
                setattr(
                    client,
                    details["operationId"],
                    generate_endpoint_method(
                        client, method, path, details["operationId"]
                    ),
                )



class ESIHubError(Exception):
    def __init__(self, message: str, status_code: int = None, details: dict = None):
        self.message = message
        self.status_code = status_code
        self.details = details or {}
        super().__init__(self.message)


class ESIHubAuthenticationError(ESIHubError):
    pass


class ESIHubRateLimitError(ESIHubError):
    pass


class ESIHubServerError(ESIHubError):
    pass


class ESIHubClientError(ESIHubError):
    pass


class ESIHubValidationError(ESIHubError):
    pass



import asyncio
from typing import Any, Dict, Optional

import redis
from cachetools import TTLCache
from multidict import CIMultiDictProxy

from .config import ESIHubConfig
from .logger import esihub_logger
from ..models import ESIHubResponse


class ESIHubCachePolicy:
    def __init__(self, ttl: int, max_size: int):
        self.ttl = ttl
        self.max_size = max_size


class ESIHubCache:
    def __init__(self, config: ESIHubConfig):
        self.config = config
        self.memory_cache = TTLCache(maxsize=1000, ttl=300)
        self.redis: Optional[redis.Redis] = None
        self.lock = asyncio.Lock()
        self.cache_enabled = self.config.get("CACHE_ENABLED", True)
        self.policies: Dict[str, ESIHubCachePolicy] = {}

    async def initialize(self):
        if not self.cache_enabled:
            esihub_logger.info("Caching is disabled.")
            return

        redis_url = self.config.get("REDIS_URL")
        if redis_url:
            try:
                self.redis = await redis.from_url(redis_url)
                await self.redis.ping()
                esihub_logger.info("Redis cache initialized successfully.")
            except Exception as e:
                esihub_logger.error(f"Failed to initialize Redis cache: {str(e)}")
                self.redis = None
        else:
            esihub_logger.info("Redis URL not provided. Using memory cache only.")

    async def get(
        self, method: str, path: str, params: Dict[str, Any]
    ) -> Optional[ESIHubResponse]:
        if not self.cache_enabled:
            return None

        cache_key = self._generate_cache_key(method, path, params)

        # Check memory cache first
        if cache_key in self.memory_cache:
            esihub_logger.debug("Cache hit (memory)", extra={"cache_key": cache_key})
            return self.memory_cache[cache_key]

        # Check Redis cache
        if self.redis:
            data = await self.redis.get(cache_key)
            if data:
                esi_response = ESIHubResponse.model_validate_json(data)
                self.memory_cache[cache_key] = esi_response
                esihub_logger.debug("Cache hit (Redis)", extra={"cache_key": cache_key})
                return esi_response

        esihub_logger.debug("Cache miss", extra={"cache_key": cache_key})
        return None

    async def set(
        self,
        method: str,
        path: str,
        params: Dict[str, Any],
        response: ESIHubResponse,
        headers: CIMultiDictProxy[str],
    ):
        if not self.cache_enabled:
            return

        cache_key = self._generate_cache_key(method, path, params)
        policy = self.get_policy(path)
        expires_in = self._get_cache_expiry(headers, policy)

        async with self.lock:
            self.memory_cache[cache_key] = response
            if self.redis:
                await self.redis.setex(
                    cache_key, expires_in, response.model_dump_json()
                )

        esihub_logger.debug(
            "Cached", extra={"cache_key": cache_key, "expires_in": expires_in}
        )

    async def invalidate(self, pattern: str):
        if not self.cache_enabled:
            return

        if self.redis:
            keys = await self.redis.keys(pattern)
            if keys:
                await self.redis.delete(*keys)
                esihub_logger.debug(
                    "Invalidated Redis cache", extra={"pattern": pattern}
                )

        # Invalidate memory cache
        for key in list(self.memory_cache.keys()):
            if pattern in key:
                del self.memory_cache[key]
        esihub_logger.debug("Invalidated memory cache", extra={"pattern": pattern})

    def _generate_cache_key(
        self, method: str, path: str, params: Dict[str, Any]
    ) -> str:
        sorted_params = sorted(params.items())
        return f"{method}:{path}:{sorted_params}"

    def _get_cache_expiry(
        self, headers: CIMultiDictProxy[str], policy: ESIHubCachePolicy
    ) -> int:
        if "Expires" in headers:
            from email.utils import parsedate
            import time

            expires = parsedate(headers["Expires"])
            if expires:
                return max(0, int(time.mktime(expires) - time.time()))
        return (
            policy.ttl
            if policy
            else int(headers.get("Cache-Control", "300").split("=")[-1])
        )

    async def close(self):
        if self.redis:
            await self.redis.close()
            esihub_logger.info("Redis connection closed.")

    def set_policy(self, path: str, policy: ESIHubCachePolicy):
        self.policies[path] = policy

    def get_policy(self, path: str) -> Optional[ESIHubCachePolicy]:
        return self.policies.get(path)



import asyncio
from typing import Any, Callable, Dict, List


class ESIHubEventSystem:
    def __init__(self):
        self.listeners: Dict[str, List[Callable]] = {}

    def on(self, event: str):
        def decorator(func: Callable):
            if event not in self.listeners:
                self.listeners[event] = []
            self.listeners[event].append(func)
            return func

        return decorator

    async def emit(self, event: str, **kwargs: Any):
        if event in self.listeners:
            await asyncio.gather(
                *(listener(**kwargs) for listener in self.listeners[event])
            )

    def remove(self, event: str, listener: Callable):
        if event in self.listeners and listener in self.listeners[event]:
            self.listeners[event].remove(listener)



import asyncio
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Type,
    TypeVar,
    Callable,
    Awaitable,
    AsyncIterator,
)

import aiohttp
from aiohttp import ClientSession, TCPConnector
from pydantic import BaseModel, ValidationError

from esihub.api.endpoints import generate_endpoints
from esihub.auth import ESIHubAuth
from esihub.core.async_profiler import ESIHubAsyncProfiler, profile
from esihub.core.background_tasks import ESIHubBackgroundTaskManager
from esihub.core.cache import ESIHubCache
from esihub.core.config import ESIHubConfig, esi_config
from esihub.core.dry_run import ESIHubDryRunMode
from esihub.core.error_handler import ESIHubErrorHandler
from esihub.core.event_system import ESIHubEventSystem
from esihub.core.logger import configure_logging, esihub_logger
from esihub.core.metrics import ESIHubMetrics
from esihub.core.rate_limiter import ESIHubRateLimiter
from esihub.exceptions import ESIHubValidationError, ESIHubError
from esihub.models import ESIHubResponse, ESIHubRequestParams
from esihub.utils import (
    validate_url,
    retry_with_exponential_backoff,
    validate_input,
    load_swagger_spec,
)

T = TypeVar("T", bound="ESIHubClient")


class ESIHubClient:
    def __init__(
        self,
        config: ESIHubConfig = esi_config,
        auth: Optional[ESIHubAuth] = None,
        cache: Optional[ESIHubCache] = None,
        rate_limiter: Optional[ESIHubRateLimiter] = None,
        error_handler: Optional[ESIHubErrorHandler] = None,
        event_system: Optional[ESIHubEventSystem] = None,
    ) -> None:
        self.config = config
        self.base_url = self.config.get("ESI_BASE_URL")
        if not validate_url(self.base_url):
            raise ValueError("Invalid base URL")
        if self.config.get("USE_HTTPS") and not self.base_url.startswith("https://"):
            raise ValueError("HTTPS is required")

        self.auth = auth or ESIHubAuth(self, config)
        self.cache = cache or ESIHubCache(config)
        self.rate_limiter = rate_limiter or ESIHubRateLimiter(config)
        self.error_handler = error_handler or ESIHubErrorHandler()
        self.event_system = event_system or ESIHubEventSystem()
        self.session: Optional[ClientSession] = None
        self.semaphore = asyncio.Semaphore(
            self.config.get("MAX_CONCURRENT_REQUESTS", 100)
        )

        self.background_tasks = ESIHubBackgroundTaskManager()
        self.metrics = ESIHubMetrics()
        self.dry_run_mode = ESIHubDryRunMode(self)

        self.profiler = ESIHubAsyncProfiler()

        self.swagger_spec = load_swagger_spec()

        configure_logging(config)
        generate_endpoints(self)

    async def __aenter__(self: T) -> T:
        await self.initialize()
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[Any],
    ) -> None:
        await self.close()

    async def initialize(self) -> None:
        ssl_context = self._create_ssl_context()
        connector = TCPConnector(
            limit=self.config.get("MAX_CONNECTIONS", 100), ssl=ssl_context
        )
        self.session = ClientSession(
            connector=connector,
            headers={"User-Agent": self.config.get("ESI_USER_AGENT")},
        )
        await self.cache.initialize()
        await self.background_tasks.start()

    async def close(self) -> None:
        if self.session:
            await self.session.close()
        await self.cache.close()
        await self.background_tasks.stop()

    @retry_with_exponential_backoff()
    @profile
    async def request(
        self,
        method: str,
        path: str,
        model: Optional[Type[BaseModel]] = None,
        **kwargs: Any,
    ) -> ESIHubResponse:
        if not validate_input(path, r"^/[\w\-/{}]+$"):
            raise ValueError("Invalid path")

        if self.config.get("DRY_RUN"):
            return await self.dry_run_mode.request(method, path, **kwargs)

        async with self.semaphore:
            with self.metrics.measure_request_duration(method, path):
                self.metrics.increment_request(method, path)
                try:
                    response = await self._make_request(method, path, **kwargs)
                    if model:
                        try:
                            validated_data = model(**response.data)
                            response.data = validated_data.model_dump()
                        except ValidationError as e:
                            raise ESIHubValidationError(
                                f"Response validation failed: {e}"
                            )
                    return response
                except Exception as e:
                    self.metrics.increment_error(type(e).__name__)
                    raise

    async def _make_request(
        self, method: str, path: str, **kwargs: Any
    ) -> ESIHubResponse:
        if not self.session:
            await self.initialize()

        url = f"{self.base_url}/latest{path}"
        await self.rate_limiter.acquire(path)

        try:
            cached_response = await self.cache.get(method, path, kwargs)
            if cached_response:
                return cached_response

            params = ESIHubRequestParams(method=method, path=path, **kwargs)
            await self.event_system.emit("before_request", params=params)

            esihub_logger.info("Making request", extra={"method": method, "path": path})

            async with self.session.request(method, url, **kwargs) as response:
                response_data = await response.json()

                self.rate_limiter.update_limit(path, response.headers)

                if response.status >= 400:
                    await self.error_handler.handle_error(
                        response.status, response_data
                    )

                esi_response = ESIHubResponse(
                    status=response.status,
                    headers=dict(response.headers),
                    data=response_data,
                )

                await self.cache.set(
                    method, path, kwargs, esi_response, response.headers
                )

                esihub_logger.info(
                    "Received response", extra={"status": response.status}
                )
                await self.event_system.emit(
                    "after_request", params=params, response=esi_response
                )

                return esi_response
        except aiohttp.ClientError as e:
            esihub_logger.error("Request failed", extra={"error": str(e)})
            raise ESIHubError(f"Request failed: {str(e)}")

    async def batch_request(self, requests: List[Dict[str, Any]]) -> tuple[Any]:
        async def bounded_request(req: Dict[str, Any]) -> ESIHubResponse:
            return await self.request(**req)

        return await asyncio.gather(*(bounded_request(req) for req in requests))

    async def stream_request(
        self, method: str, path: str, **kwargs: Any
    ) -> AsyncIterator[Any]:
        if not self.session:
            await self.initialize()

        url = f"{self.base_url}{path}"
        await self.rate_limiter.acquire(path)

        async with self.session.request(method, url, **kwargs) as response:
            async for chunk in response.content.iter_any():
                yield chunk

    async def bulk_operation(
        self, operation: str, items: List[Any], batch_size: int = 100
    ) -> List[Any]:
        results = []
        for i in range(0, len(items), batch_size):
            batch = items[i : i + batch_size]
            batch_results = await getattr(self, operation)(batch)
            results.extend(batch_results)
        return results

    def _create_ssl_context(self):
        import ssl

        ssl_context = ssl.create_default_context()
        ssl_context.minimum_version = ssl.TLSVersion.TLSv1_2
        return ssl_context

    async def paginated_request(
        self, method: str, path: str, **kwargs
    ) -> AsyncIterator[Dict[str, Any]]:
        page = 1
        while True:
            kwargs["params"] = kwargs.get("params", {})
            kwargs["params"]["page"] = page
            response = await self.request(method, path, **kwargs)
            yield response.data
            if "X-Pages" in response.headers and page < int(
                response.headers["X-Pages"]
            ):
                page += 1
            else:
                break

    async def add_background_task(
        self, coroutine: Callable[..., Awaitable[Any]], *args, **kwargs
    ) -> None:
        await self.background_tasks.add_task(coroutine, *args, **kwargs)

    def __getattr__(self, name: str):
        if (
            name.startswith("get_")
            or name.startswith("post_")
            or name.startswith("put_")
            or name.startswith("delete_")
        ):
            method, *path_parts = name.split("_")
            path = "/" + "/".join(path_parts) + "/"

            async def dynamic_endpoint(**kwargs):
                return await self.request(method.upper(), path, **kwargs)

            return dynamic_endpoint
        raise AttributeError(
            f"'{self.__class__.__name__}' object has no attribute '{name}'"
        )



import os
from typing import Any, Dict


class ESIHubConfig:
    def __init__(self):
        self.config: Dict[str, Any] = {
            "ESI_CLIENT_ID": os.getenv("ESI_CLIENT_ID"),
            "ESI_CLIENT_SECRET": os.getenv("ESI_CLIENT_SECRET"),
            "ESI_CALLBACK_URL": os.getenv("ESI_CALLBACK_URL"),
            "ESI_BASE_URL": os.getenv("ESI_BASE_URL", "https://esi.evetech.net"),
            "ESI_USER_AGENT": os.getenv("ESI_USER_AGENT", "ESIHub/1.0"),
            "ESI_RETRY_ATTEMPTS": int(os.getenv("ESI_RETRY_ATTEMPTS", "3")),
            "ESI_RATE_LIMIT": int(os.getenv("ESI_RATE_LIMIT", "150")),
            "REDIS_URL": os.getenv("REDIS_URL"),
            "LOG_LEVEL": os.getenv("LOG_LEVEL", "INFO"),
            "USE_HTTPS": os.getenv("USE_HTTPS", "True").lower() == "true",
            "MAX_CONCURRENT_REQUESTS": int(os.getenv("MAX_CONCURRENT_REQUESTS", "100")),
            "MAX_CONNECTIONS": int(os.getenv("MAX_CONNECTIONS", "100")),
            "DRY_RUN": os.getenv("DRY_RUN", "False").lower() == "true",
        }

    def get(self, key: str, default: Any = None) -> Any:
        return self.config.get(key, default)

    def set(self, key: str, value: Any) -> None:
        self.config[key] = value

    def update(self, config: Dict[str, Any]) -> None:
        self.config.update(config)


esi_config = ESIHubConfig()



from typing import Any, Dict

from .logger import esihub_logger
from ..models import ESIHubResponse


class ESIHubDryRunMode:
    def __init__(self, client):
        self.client = client

    async def request(self, method: str, path: str, **kwargs: Any) -> ESIHubResponse:
        esihub_logger.info(f"Dry run request: {method} {path}")
        esihub_logger.info(f"Request parameters: {kwargs}")

        # Validate the request parameters here
        self._validate_params(kwargs)

        # Simulate a successful response with echo
        mock_response = ESIHubResponse(
            status=200,
            headers={"Content-Type": "application/json"},
            data=kwargs,
        )

        esihub_logger.info(f"Dry run response: {mock_response}")
        return mock_response

    def _validate_params(self, params: Dict[str, Any]) -> None:
        # Add any parameter validation logic here
        # For example, check if required parameters are present
        required_params = self._get_required_params(params.get("path", ""))
        for param in required_params:
            if param not in params:
                esihub_logger.warning(f"Missing required parameter: {param}")

    def _get_required_params(self, path: str) -> list:
        # This method should return a list of required parameters for the given path
        # You would typically implement this based on the ESI API specification
        return []
